import copy
import os

import cv2
import numpy as np
import tqdm

from kano.file_utils import create_folder
from kano.image_utils import concatenate_images


def get_frame_at_second(video_path, target_second):
    """
    Get numpy array of a frame from a video at the given second

    Args:
        video_path (str): path of the video
        target_second (int): second timestamp to get frame

    Returns:
        frame (np.ndarray): numpy array of the frame at the given second
    """
    cap = cv2.VideoCapture(video_path)

    if not cap.isOpened():
        raise ValueError("Video file could not be opened.")

    fps = cap.get(cv2.CAP_PROP_FPS)
    target_frame = int(target_second * fps)

    cap.set(cv2.CAP_PROP_POS_FRAMES, target_frame)

    ret, frame = cap.read()

    cap.release()

    if ret:
        return frame
    else:
        raise ValueError("Frame not found at the specified second.")


def extract_frames(video_path, target_folder, seconds_interval):
    """
    Extract frames from a video with a given seconds interval

    Args:
        video_path (str): path of the video
        target_folder (str): path to save extracted frames
        seconds_interval (float): amount of seconds between two extracted frames
    """
    cap = cv2.VideoCapture(video_path)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    video_fps = cap.get(cv2.CAP_PROP_FPS)
    frame_interval = int(video_fps * seconds_interval)

    create_folder(target_folder)

    max_length = len(str(frame_count))
    for frame_number in tqdm.tqdm(range(0, frame_count, frame_interval)):
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)
        ret, frame = cap.read()
        if not ret:
            break
        number = str(frame_number).zfill(max_length)
        frame_filename = os.path.join(target_folder, f"frame_{number}.jpg")
        cv2.imwrite(frame_filename, frame)

    cap.release()


def cut_video(input_video_path, output_video_path, start_second, end_second):
    """
    Cut a segment from a video file and save it as a new video.

    Args:
        input_video_path (str): Path to the input video file.
        output_video_path (str): Path to save the output video file.
        start_second (float): Start time of the segment to be cut in seconds.
        end_second (float): End time of the segment to be cut in seconds.
    """
    cap = cv2.VideoCapture(input_video_path)
    fps = cap.get(cv2.CAP_PROP_FPS)
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    start_frame = int(start_second * fps)
    end_frame = int(end_second * fps)

    fourcc = cv2.VideoWriter_fourcc(*"XVID")
    out = cv2.VideoWriter(output_video_path, fourcc, fps, (width, height))

    cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
    while cap.isOpened() and cap.get(cv2.CAP_PROP_POS_FRAMES) <= end_frame:
        ret, frame = cap.read()
        if not ret:
            break
        out.write(frame)

    cap.release()
    out.release()


def add_title(
    image, title, font_scale=2, font_thickness=3, title_padding_size=10
):
    font = cv2.FONT_HERSHEY_SIMPLEX
    text_size = cv2.getTextSize(title, font, font_scale, font_thickness)[0]
    text_x = (image.shape[1] - text_size[0]) // 2
    text_y = image.shape[0] + title_padding_size + text_size[1]
    height_diff = title_padding_size + text_size[1] + title_padding_size
    padding = np.zeros((height_diff, image.shape[1], 3), dtype=np.uint8)
    padded_image = np.vstack([image, padding])
    cv2.putText(
        padded_image,
        title,
        (text_x, text_y),
        font,
        font_scale,
        (255, 255, 255),
        font_thickness,
    )
    return padded_image


def get_video_duration(video_path):
    cap = cv2.VideoCapture(video_path)
    fps = cap.get(cv2.CAP_PROP_FPS)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    duration = frame_count / fps
    cap.release()
    return duration


def concatenate_videos(
    video_paths,
    titles=None,
    output_video_path="concatenated_video.mp4",
    total_seconds=None,
    font_scale=2,
    font_thickness=3,
    title_padding_size=10,
    frame_padding_size=10,
):
    """
    Concatenate multiple video files into a single video.

    Args:
        video_paths (list): A 2D list of video file paths to be concatenated.
        titles (list): A 2D list of titles corresponding to each video segment.
        output_video_path (str): Path to save the concatenated video file.
        total_seconds (float): Total duration of the output video in seconds.
        font_scale (int): Font scale for titles.
        font_thickness (int): Font thickness for titles.
        title_padding_size (int): Padding size for titles.
        frame_padding_size (int): Padding size between frames.
    """
    # ensure video_paths is a 2D list
    new_video_paths = copy.deepcopy(video_paths)
    if not isinstance(video_paths[0], list):
        new_video_paths = [copy.deepcopy(video_paths)]

    rows = len(new_video_paths)
    cols = max(len(row) for row in new_video_paths)

    for video_path in new_video_paths:
        video_path += [None] * (cols - len(video_path))

    # get target video duration
    video_captures = list()
    durations = list()

    max_width = 0
    max_height = 0
    sample_cap = None
    for i, row in enumerate(new_video_paths):
        video_captures.append(list())
        for video_path in row:
            if video_path is not None:
                sample_cap = cv2.VideoCapture(video_path)
                width = int(sample_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                height = int(sample_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                max_width = max(width, max_width)
                max_height = max(height, max_height)
                video_captures[i].append(sample_cap)
                durations.append(get_video_duration(video_path))
            else:
                video_captures[i].append(None)

    min_duration = min(durations)
    if total_seconds is not None:
        min_duration = min(min_duration, total_seconds)

    # init output video
    fps = int(sample_cap.get(cv2.CAP_PROP_FPS))
    fourcc = cv2.VideoWriter_fourcc(*"XVID")

    font = cv2.FONT_HERSHEY_SIMPLEX
    text_size = cv2.getTextSize("SAMPLE", font, font_scale, font_thickness)[0]

    output_width = max_width * cols + frame_padding_size * (cols - 1)
    output_height = (
        max_height + (text_size[1] + title_padding_size * 2)
    ) * rows + frame_padding_size * (rows - 1)

    output_video = cv2.VideoWriter(
        output_video_path, fourcc, fps, (output_width, output_height)
    )
    num_frames = int(min_duration * fps)

    for __ in tqdm.tqdm(range(num_frames), desc="Videos concatenating"):
        frames = list()
        for row in range(rows):
            frames.append(list())
            for col in range(cols):
                cap = video_captures[row][col]
                if (
                    len(titles) < row + 1
                    or len(titles[row]) < col + 1
                    or titles[row][col] is None
                ):
                    title = " "
                else:
                    title = titles[row][col]

                frame = np.zeros((max_height, max_width, 3), dtype=np.uint8)
                if cap is not None:
                    __, frame = cap.read()

                frame = add_title(
                    frame,
                    title,
                    font_scale,
                    font_thickness,
                    title_padding_size,
                )

                frames[row].append(frame)

        concatenated_frame = concatenate_images(frames, frame_padding_size)

        output_video.write(concatenated_frame)

    for row_caps in video_captures:
        for cap in row_caps:
            if cap is not None:
                cap.release()

    output_video.release()



import cv2
import numpy as np


def extract_bbox_area(image, bbox):
    """
    Return cropped image from the given bounding box area

    Args:
        image (np.array) with shape (H, W, 3): image to extract the box
        bbox (np.array) with shape (4,): xyxy location of the box

    Returns:
        cropped_image (np.array): with shape (new_H, new_W, 3) based on bbox
    """

    (left, top), (right, bottom) = bbox[:2], bbox[2:]
    return image.copy()[top:bottom, left:right]


def xywh2xyxy(xywh):
    """
    Converts bounding box coordinates from (x_center, y_center, width, height) format to (x_min, y_min, x_max, y_max) format.

    Args:
        xywh (np.array) with shape (4,): A tuple containing (x_center, y_center, width, height) of the bounding box.

    Returns:
        xyxy (tuple(int)): xyxy location of the bounding box.
    """

    x_center, y_center, bbox_width, bbox_height = xywh
    x_min = int(x_center - bbox_width / 2)
    y_min = int(y_center - bbox_height / 2)
    x_max = int(x_center + bbox_width / 2)
    y_max = int(y_center + bbox_height / 2)

    return x_min, y_min, x_max, y_max


def draw_bbox(
    image, bbox, bbox_type="xyxy", bbox_color=(0, 0, 255), label=None
):
    """
    Draws a bounding box on the image.

    Args:
        image (np.ndarray or str): The image on which the bounding box will be drawn.
        bbox (list or np.ndarray): The bounding box coordinates. If it's a list, it should be in the format specified by bbox_type.
        bbox_type (str): Type of bounding box coordinates. Should be either "xyxy" or "xywh" or "s_xywh".
        bbox_color (tuple): Color of the bounding box in BGR format.
        label (str, optional): Label to be displayed alongside the bounding box.

    Returns:
        temp_image (np.ndarray): Image with the bounding box drawn.
    """
    if isinstance(image, str):
        temp_image = cv2.imread(image)
    else:
        temp_image = image.copy()

    image_height, image_width = temp_image.shape[:2]

    temp_bbox = bbox.copy()
    if isinstance(bbox, list):
        temp_bbox = np.array(bbox)

    if "s_" in bbox_type:
        temp_bbox *= np.array(
            [image_width, image_height, image_width, image_height]
        )

    temp_bbox = temp_bbox.astype(np.int64)

    if "xyxy" in bbox_type:
        x_min, y_min, x_max, y_max = temp_bbox
    elif "xywh" in bbox_type:
        x_min, y_min, x_max, y_max = xywh2xyxy(temp_bbox)
    else:
        raise ValueError("Invalid bounding box type")

    cv2.rectangle(temp_image, (x_min, y_min), (x_max, y_max), bbox_color, 2)
    if label is not None:
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 1
        thickness = 2
        (text_width, text_height), _ = cv2.getTextSize(
            label, font, font_scale, thickness
        )

        background_position = (x_min, y_min)
        background_end_position = (
            x_min + text_width,
            y_min - text_height - 5,
        )
        cv2.rectangle(
            temp_image,
            background_position,
            background_end_position,
            bbox_color,
            -1,
        )
        cv2.putText(
            temp_image,
            label,
            (x_min, y_min),
            font,
            font_scale,
            (255, 255, 255),
            thickness,
        )

    return temp_image


def calculate_iou(xyxy1, xyxy2):
    x1 = max(xyxy1[0], xyxy2[0])
    y1 = max(xyxy1[1], xyxy2[1])
    x2 = min(xyxy1[2], xyxy2[2])
    y2 = min(xyxy1[3], xyxy2[3])

    intersection_area = max(0, x2 - x1 + 1) * max(0, y2 - y1 + 1)

    area_box1 = (xyxy1[2] - xyxy1[0] + 1) * (xyxy1[3] - xyxy1[1] + 1)
    area_box2 = (xyxy2[2] - xyxy2[0] + 1) * (xyxy2[3] - xyxy2[1] + 1)

    union_area = area_box1 + area_box2 - intersection_area

    iou = intersection_area / union_area

    return iou


class FailedDetectionTypes:
    UnmatchedLabels = 0
    Nothing = 1
    BelowIoUThreshold = 2

    message_dict = {
        0: "Unmatched number of labels",
        1: "Everything is ok",
        2: "IoU between Groundtruth bbox and Predicted bbox is below the given threshold",
    }

    @classmethod
    def print_message(cls, id):
        print(cls.message_dict[id])


def get_failed_detected_results(labels, predictions, iou_threshold=0.5):
    if len(labels) != len(predictions):
        return FailedDetectionTypes.UnmatchedLabels

    checked_prediction_indicies = list()

    for label in labels:
        index = 0
        max_iou = 0
        for i, pred in enumerate(predictions):
            if (
                pred["class"] == label["class"]
                and i not in checked_prediction_indicies
            ):
                iou = calculate_iou(label["xyxy"], pred["xyxy"])
                if iou > max_iou:
                    max_iou = iou
                    index = i

        if index in checked_prediction_indicies:
            return FailedDetectionTypes.UnmatchedLabels
        if max_iou < iou_threshold:
            return FailedDetectionTypes.BelowIoUThreshold

        checked_prediction_indicies.append(index)

    return FailedDetectionTypes.Nothing
